package backend;

import intermediate.SymbolTable;
import intermediate.SymtabEntry;
import intermediate.SymtabEntryType;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

/**
 * The runtime activation record.
 */
public class ActivationRecord 
{
    private final int level;                       // current nesting level
    private TreeMap<String, Object> store = null;  // store of name-value pairs
    private ActivationRecord link = null;          // back link
    
    /**
     * Constructor. Initialize the store by creating slots
     * indicated by the scope's symbol table.
     * @param symtab the symbol table.
     */
    public ActivationRecord(SymbolTable symtab)
    {
        this.level = symtab.getLevel();
        this.link  = null;
        this.store = new TreeMap<>();
        
        Iterator<Map.Entry<String, SymtabEntry>> it = symtab.iterator();
        
        // Iterate over the symbol table entries.
        while (it.hasNext()) {
            Map.Entry<String, SymtabEntry> entry = it.next();
            String name = entry.getKey();
            
            // For each variable name in the symbol table, create a slot.
            if (entry.getValue().getType() == SymtabEntryType.VARIABLE) {
                store.put(name, null);
            }
        }
    }
    
    public int              getLevel() { return this.level; }
    public ActivationRecord getLink()  { return this.link; }
    
    public void setLink(ActivationRecord to) { this.link = to; }
    
    /**
     * Get a variable's value from the activation record.
     * @param name the variable's name.
     * @return the value.
     */
    public Object getValue(String name)
    {
        return store.get(name);
    }
    
    /**
     * Set a variable's value in the activation record.
     * @param name the variable's name.
     * @param value the value to set.
     */
    public void setValue(String name, Object value)
    {
        store.put(name, value);
    }
}
