package backend;

import java.util.ArrayList;

import frontend.Token;
import frontend.TokenType;
import intermediate.Node;
import intermediate.NodeType;
import intermediate.ParserResults;
import intermediate.SymbolTable;
import intermediate.SymtabEntry;
import intermediate.SymtabEntryType;

/**
 * The limited Scheme interpreter's back-end executor.
 * Uses results from the front-end parser.
 */
public class Executor 
{
    private ArrayList<Node> trees;       // top-level parse trees
    private SymbolTable symtabTopLevel;  // top-level symbol table
    
    /**
     * Constructor.
     * @param parserResults the results from the parser.
     */
    public Executor(ParserResults parserResults)
    {
        this.trees = parserResults.getTrees();
        this.symtabTopLevel = parserResults.getSymtabTopLevel();
    }
    
    /**
     * Execute the top-level list expressions.
     */
    public void execute()
    {
        System.out.println("\n===== BEGIN EXECUTION =====\n");
        
        ActivationRecord ar1 = new ActivationRecord(symtabTopLevel);
        RuntimeStack.push(ar1);
        
        for (Node root : trees) {
            Node car = root.car();
            Token token = car.getToken();
            
            // Execute a define to set a variable's value.
            if (token.getType() == TokenType.KW_DEFINE) {
                Node cddr  = root.cdr().cdr();
                Node cddar = cddr.car();
                
                if (cddar != null) {
                    
                    // The value is an element.
                    if (cddar.getType() == NodeType.ELEMENT) {
                        Object value = executeDefine(root);
                        print(value);
                    }
                    else {
                        Node cddaar = cddar.car();
                        
                        // The value is a list other than lambda.
                        if (cddaar != null) {
                            token = cddaar.getToken();
                            if (token.getType() != TokenType.KW_LAMBDA) {
                                Object value = executeDefine(root);
                                print(value);
                            }
                        }
                    }
                }
            }
            
            // Not a define. Just evaluate the list expression.
            else {
                Object value = eval(root);
                print(value);
            }
        }
    }
    
    /**
     * Evaluate an expression.
     * @param root the root of the expression tree.
     * @return the expression's value.
     */
    private Object eval(Node root)
    {
        NodeType nodeType = root.getType();
        
        if (nodeType == NodeType.ELEMENT) {
            return evalElement(root);
        }
        else {
            return evalList(root);
        }
    }
    
    /**
	 * Evaluate an element.
	 * @param element the element's tree node.
	 * @return the element's value.
	 */
	private Object evalElement(Node element)
	{
	    Token token = element.getToken();
	    TokenType tokenType = token.getType();
	    
	    if (tokenType == TokenType.NUMBER) {
	        return token.getValue();
	    }
	    else if (tokenType == TokenType.IDENTIFIER) {
	        SymtabEntry entry = element.getSymtabEntry();
	        return RuntimeStack.getValue(entry);
	    }
	    else {
	        return null;
	    }
	}

	/**
	 * Evaluate a list expression.
	 * @param root the root of the list expression tree.
	 * @return the list's value.
	 */
	private Object evalList(Node root)
	{
	    Node car = root.car();
	    Token token = car.getToken();
	    
	    // If it's a let or let* expression, execute it.
	    if (   (token.getType() == TokenType.KW_LET)
	        || (token.getType() == TokenType.KW_LET_STAR)) {
	        return executeLet(root);
	    }
	    
	    // Execute a procedure call.
	    else {
	        SymtabEntry procEntry = car.getSymtabEntry();
	        SymtabEntryType procType = procEntry.getType();
	        Token procToken = car.getToken();
	        
	        // Create an array list of actual argument values.
	        ArrayList<Object> args = new ArrayList<>();
	        for (Node n = root.cdr(); n != null; n = n.cdr()) {
	            car = n.car();
	            args.add(eval(car));
	        }
	        
	        // Call a built-in or a defined procedure.
	        if (procType == SymtabEntryType.BUILT_IN) {
	            return BuiltIn.eval(procToken, args);
	        }
	        else {
	            return executeCall(root);
	        }
	    }
	}

	/**
	 * Execute a define.
	 * @param defineRoot the root of the define subtree.
	 * @return the define's value.
	 */
	private Object executeDefine(Node defineRoot)
	{
	    Node cdar = defineRoot.cdr().car();
	    SymtabEntry entry = cdar.getSymtabEntry();
	    
	    Node cddar = defineRoot.cdr().cdr().car();
	    Token token = cddar.getToken();
	    Number value = token.getValue();
	    
	    return RuntimeStack.setValue(entry, value);
	}

	/**
     * Execute a procedure call.
     * @param callRoot the root of the call tree.
     * @return the value of the procedure call.
     */
    private Object executeCall(Node callRoot)
    {
        Node car = callRoot.car();
        
        // Get the root of the procedure's lambda tree
        // from the procedure name's symbol table entry.
        SymtabEntry procEntry = car.getSymtabEntry();        
        Node lambdaRoot = procEntry.getNode();
        
        // Execute the lambda expression and return its value.
        Object value = executeLambda(callRoot, lambdaRoot);
        return value;
    }
    
    /**
     * Execute a lambda expression.
     * @param callRoot the root of the call subtree.
     * @param lambdaRoot the root of the lambda subtree.
     * @return the value of the lambda expression.
     */
    private Object executeLambda(Node callRoot, Node lambdaRoot)
    {
        // Create the activation record for the lambda's scope.
        SymbolTable lambdaSymtab = lambdaRoot.getSymtab();
        ActivationRecord record = new ActivationRecord(lambdaSymtab);
        
        Node actuals = callRoot.cdr();          // root of the actual parms
        Node formals = lambdaRoot.cdr().car();  // root of the formal parms
        
        // Evaluate the actual parameters and set the formal parameter values.
        for (Node a = actuals, f = formals; a != null; a = a.cdr(), f = f.cdr()) {
            String name = f.car().getToken().getText();
            Object value = eval(a.car());
            record.setValue(name, value);
        }
        
        // Now push the lambda's activation record onto the runtime stack.
        RuntimeStack.push(record);
        
        // Evaluate the lambda expression's body.
        Node body = lambdaRoot.cdr().cdr().car();
        Object value = eval(body);
        
        // Pop off the activation record and return the computed value.
        RuntimeStack.pop();
        return value;
    }
    
    /**
     * Evaluate a let or let* expression.
     * @param letRoot the root of the let subtree.
     * @return the value of the let expression.
     */
    private Object executeLet(Node letRoot)
    {
        // Create an activation record for the let's scope
        // and push it onto the runtime stack.
        SymbolTable letSymtab = letRoot.getSymtab();
        ActivationRecord record = new ActivationRecord(letSymtab);
        RuntimeStack.push(record);

        // Evaluate and set the value of each variable.
        Node varList = letRoot.cdr().car();
        for (Node v = varList; v != null; v = v.cdr()) {
            Node pair = v.car();
            String name = pair.car().getToken().getText();
            Object value = eval(pair.cdr().car());
            record.setValue(name, value);
        }
        
        // Evaluate the let expression's body.
        Node body = letRoot.cdr().cdr().car();
        Object value = eval(body);
        
        // Pop off the activation record and return the computed value.
        RuntimeStack.pop();
        return value;
    }
    
    /**
     * Print a computed value, a scalar or a list.
     * @param value the value to print.
     */
    private void print(Object value)
    {
        if (value instanceof Number) {
            System.out.println(value);
        }
        else if (value instanceof Node) {
            TreePrinter.print((Node) value);
        }
    }
}
