package frontend;

import intermediate.Node;
import intermediate.ParserResults;
import intermediate.SymbolTable;
import intermediate.SymtabStack;
import intermediate.TreeWalker;

import java.util.ArrayList;


import backend.TreePrinter;

/**
 * A limited Scheme parser.
 */
public class Parser
{
    private Scanner scanner;
    private ArrayList<Node> trees;
    
    /**
     * Constructor.
     * @param scanner the simple Scheme scanner.
     */
    public Parser(Scanner scanner)
    {
        this.scanner = scanner;
        this.trees = new ArrayList<Node>();
    }
    /**
     * The parse method.
     */
    public ParserResults parse()
    {
        Token token;
        
        // Global symbol table.
        SymbolTable symtabGlobal = SymbolTable.createGlobal();
        SymtabStack.push(symtabGlobal);
        
        // Top-level symbol table.
        SymbolTable symtabTopLevel = new SymbolTable(1);        
        SymtabStack.push(symtabTopLevel);
        
        TreeWalker walker = new TreeWalker(symtabTopLevel);
        
        // Loop to get tokens until the end of file.
        while ((token = scanner.nextToken()).getType() != TokenType.EOF) {
            TokenType tokenType = token.getType();
            
            if (tokenType == TokenType.SS_LPAREN) {
                Node root = parseList(1);
                trees.add(root);
                
                // Walk the tree to set up symbol tables.
                walker.walk(root, 1);
            }
        }
        
        // Print the global and top-level symbol tables.
        symtabGlobal.print();
        symtabTopLevel.print();
        
        SymtabStack.pop();
        
        // Print the parse trees.
        for (Node tree : trees) TreePrinter.print(tree);
        
        // Return the results of the parse.
        return new ParserResults(symtabTopLevel, trees);
    }
    
    /**
     * Parse a list and build a parse tree.
     * @param level the nesting level
     * @return the root of the tree.
     */
    private Node parseList(int level)
    {
        Node root = new Node();
        Node currentNode = null;
        
        // Get the first token after the opening left parenthesis.
        Token token = scanner.nextToken();
        TokenType tokenType = token.getType();
        
        // Loop to get tokens until the closing right parenthesis.
        while (tokenType != TokenType.SS_RPAREN) {
            
            // Set currentNode initially to the root,
            // then move it down the cdr links.
            if (currentNode == null) {
                currentNode = root;
            }
            else {
                Node newNode = new Node();
                currentNode.setCdr(newNode);
                currentNode = newNode;
            }
            
            // Left parenthesis: Parse a sublist and return the root
            // of the subtree which is set as the car of the current node.
            // Otherwise, set the token as the data of the current node.
            if (tokenType == TokenType.SS_LPAREN) {
                currentNode.setCar(parseList(level+1));
            }
            else {
                currentNode.setCar(new Node(token));
            }
            
            // Get the next token for the next time around the loop.
            token = scanner.nextToken();
            tokenType = token.getType();
        }
        
        return root;  // of the parse tree
    }
}
