package intermediate;

import frontend.Token;
import frontend.TokenType;

/**
 * Parse tree walker to set up the symbol tables.
 */
public class TreeWalker 
{
    private SymbolTable symtabTopLevel;  // top-level symbol table
    
    /**
     * Constructor.
     * @param symtabTopLevel the top-level symbol table.
     */
    public TreeWalker(SymbolTable symtabTopLevel)
    {
        this.symtabTopLevel = symtabTopLevel;
    }
    
    /**
     * Walk the parse tree to set up the symbol tables.
     * @param node the node to start walking.
     * @param level the current nesting level.
     */
    public void walk(Node node, int level)
    {
        while (node != null) {
            
            // Process an element.
            if (node.getType() == NodeType.ELEMENT) {
                lookup(node);
            }
            
            // Process a list.
            else {
                Node car = node.car();
                
                // Check to see if the list is a define.
                if (car != null) {
                    if (car.getType() == NodeType.ELEMENT) {
                        TokenType carType = car.getToken().getType();
    
                        // The car of the list is "define".
                        if (carType == TokenType.KW_DEFINE) {
                            processDefine(node, level);
                            return;
                        }
                        
                        // The car of the list is just an identifier.
                        else {
                            Token token = car.getToken();
                            if (token.getType() != TokenType.NUMBER) {
                                lookup(car);
                            }
                        }
                    }
                    
                    // The car of the list is a sublist, so walk it.
                    else {
                        walk(car, level);
                    }
                }
            }
            
            node = node.cdr();
        }
    }
    
    /**
     * Process a define.
     * @param defineRoot the root node of the define.
     * @param level the current nesting level.
     */
    private void processDefine(Node defineRoot, int level)
    {
        Node cdar = defineRoot.cdr().car();
        Node cddar = defineRoot.cdr().cdr().car();
        Token token = cdar.getToken();
        
        // The defined symbol could be a variable or a procedure.
        SymtabEntryType entryType = cddar.getType() == NodeType.ELEMENT
                                        ? SymtabEntryType.VARIABLE
                                        : SymtabEntryType.PROCEDURE;
        SymtabEntry entry = symtabTopLevel.enter(token.getText(), entryType);
        
        // Cross link: The tree node points to the symbol table entry
        //             and the symbol table entry points to the parse tree.
        cdar.setSymtabEntry(entry);
        entry.setNode(cddar);
        
        // If a procedure was defined, process its lambda expression.
        if (entryType == SymtabEntryType.PROCEDURE) {
            processLambda(cddar, level+1);
        }
    }
    
    /**
     * Process a lambda expression.
     * @param lambdaRoot the root of the lambda subtree.
     * @param level the current nesting level.
     */
    private void processLambda(Node lambdaRoot, int level)
    {
        // Create a symbol table for the lambda's scope
        // and push it onto the symbol table stack.
        SymbolTable symtab = new SymbolTable(level);
        SymtabStack.push(symtab);
        
        // Link the lambda root to the symbol table.
        lambdaRoot.setSymtab(symtab);
        
        // Enter the lambda's formal parameters into the symbol table.
        Node formals = lambdaRoot.cdr().car();
        for (Node f = formals; f != null; f = f.cdr()) {
            String varName = f.car().getToken().getText();
            symtab.enter(varName, SymtabEntryType.VARIABLE);
        }
        symtab.print();
        
        boolean letFlag = false;
        Node body = lambdaRoot.cdr().cdr();
        
        // Process the lambda body. It could be a let or let*.
        if (body != null) {
            Node cddar = body.car();
            if (cddar != null) {
                if (cddar.getType() == NodeType.LIST) {
                    Node cddaar = cddar.car();
                    if (cddaar != null) {
                        Token token = cddaar.getToken();
                        
                        // Process a let or let* expression.
                        if (   (token != null) 
                            && (   (token.getType() == TokenType.KW_LET)
                                || (token.getType() == TokenType.KW_LET_STAR))) {
                            processLet(cddar, level+1);
                            letFlag = true;
                        }
                    }
                }
            }
        }
        
        // It wasn't a let or let*. Just walk the body.
        if (!letFlag && (body != null)) {
            walk(body, level);
        }
        
        // Pop off the lambda's symbol table.
        SymtabStack.pop();
    }
    
    /**
     * Process a let or let* expression.
     * @param letRoot 
     * @param level
     */
    private void processLet(Node letRoot, int level)
    {
        // Create a symbol table for the let's scope
        // and push it onto the symbol table stack.
        SymbolTable symtab = new SymbolTable(level);
        SymtabStack.push(symtab);
        
        // Link the let root to the symbol table.
        letRoot.setSymtab(symtab);
        
        // Enter the let's variables into the symbol table.
        Node variables = letRoot.cdr().car();
        for (Node v = variables; v != null; v = v.cdr()) {
            Node varRoot = v.car();
            String varName = varRoot.car().getToken().getText();
            Node valueRoot = varRoot.cdr();
            symtab.enter(varName, SymtabEntryType.VARIABLE);
            
            // Walk the variable's value.
            walk(valueRoot, level);
        }
        symtab.print();
        
        // Walk the let's body.
        Node body = letRoot.cdr().cdr();
        walk(body, level);
        
        // Pop off the let's symbol table.
        SymtabStack.pop();
    }
    /**
     * Look up a symbol in the symbol table stack.
     * @param symbolNode the node containing the symbol.
     * @return the symbol's symbol table entry if found, else null.
     */
    private SymtabEntry lookup(Node symbolNode)
    {
        String text = symbolNode.getToken().getText();
        SymtabEntry entry = SymtabStack.find(text);
        
        if (entry != null) {
            symbolNode.setSymtabEntry(entry);
        }
        else {
            System.out.printf("*** Undefined symbol '%s'\n", text);
        }
        
        return entry;
    }
}
